# -*- coding: utf-8 -*-
from flask_alchy import Alchy

from housekeeper.store import BaseHandler, models


class HousekeeperAlchy(Alchy, BaseHandler):
    pass


store = HousekeeperAlchy(Model=models.Model)



# -*- coding: utf-8 -*-
import os
import datetime
from copy import deepcopy

from pathlib import Path
import tempfile

from housekeeper.include import include_version

def test_delete_both(store, bundle_data, bundle_data_old):
    """
    test deletion where all files are older than before date
    """
    bundle_obj, version_obj = store.add_bundle(data=bundle_data)
    store.add_commit(bundle_obj)
    bundle_old_obj, version_old_obj = store.add_bundle(data=bundle_data_old)
    store.add_commit(bundle_old_obj)

    query = store.files_before(before=str(datetime.datetime.now()))
    assert len(query.all()) == 4


def test_delete_one(store, bundle_data, bundle_data_old):
    """
    test deletion where not all files are older than before date
    """
    bundle_obj, version_obj = store.add_bundle(data=bundle_data)
    store.add_commit(bundle_obj)
    bundle_old_obj, version_old_obj = store.add_bundle(data=bundle_data_old)
    store.add_commit(bundle_old_obj)

    query = store.files_before(before="2018-02-01")
    assert len(query.all()) == 2


def test_delete_none(store, bundle_data, bundle_data_old):
    """
    test deletion where no files are older than before date
    """
    bundle_obj, version_obj = store.add_bundle(data=bundle_data)
    store.add_commit(bundle_obj)
    bundle_old_obj, version_old_obj = store.add_bundle(data=bundle_data_old)
    store.add_commit(bundle_old_obj)

    query = store.files_before(before="2017-02-01")
    assert len(query.all()) == 0


def test_delete_notondisk(store, tmpdir, bundle_data):
    """
    test deletion of files that are not on disk
    """
    with tempfile.NamedTemporaryFile(delete=True) as file1:

        bundle_data_notondisk = deepcopy(bundle_data)
        bundle_data_notondisk['files'][0]['path']=file1.name
        bundle_data_notondisk['files'][1]['path']=bundle_data_notondisk['files'][1]['path'].replace('.vcf',
                                                                                    '.2.vcf')
        bundle_data_notondisk['name']='angrybird'

        bundle_obj, version_obj = store.add_bundle(data=bundle_data)
        store.add_commit(bundle_obj)
        include_version(tmpdir, version_obj, hardlink=False)
        bundle_obj_notondisk, version_obj_notondisk = store.add_bundle(data=bundle_data_notondisk)
        store.add_commit(bundle_obj_notondisk)
        include_version(tmpdir, version_obj_notondisk, hardlink=False)

    query = store.files_before()

    assert len(query.all()) == 4

    files_notondisk = set(query) - store.files_ondisk(query)

    assert len(files_notondisk) == 1



# -*- coding: utf-8 -*-
from datetime import timedelta

PIPELINES = ('mip',)
# three months before clean up by default
TIME_TO_CLEANUP = timedelta(days=(30 * 3))
ARCHIVE_TYPES = ('data', 'result', 'meta', 'archive')
EXTRA_STATUSES = ['coverage', 'frequency', 'genotype', 'visualizer',
                  'rawdata', 'qc']



# -*- coding: utf-8 -*-
import datetime
import tempfile
import dateutil

import pytest

from housekeeper import include
from housekeeper.store import models, Store


@pytest.fixture
def version(tmpdir):
    file_path_1 = tmpdir.join('example.vcf.gz')
    file_path_1.write('content')
    file_path_1_checksum = include.checksum(file_path_1)
    file_path_2 = tmpdir.join('example2.txt')
    file_path_2.write('content')
    bundle_obj = models.Bundle(name='privatefox')
    version_obj = models.Version(created_at=datetime.datetime.now(), bundle=bundle_obj)
    version_obj.files.append(
        models.File(path=file_path_1, to_archive=True, tags=[models.Tag(name='vcf-gz')],
            checksum=file_path_1_checksum),
    )
    version_obj.files.append(
        models.File(path=file_path_2, to_archive=False, tags=[models.Tag(name='tmp')])
    )
    return version_obj


@pytest.fixture(scope='function')
def bundle_data():
    data = {
        'name': 'sillyfish',
        'created': datetime.datetime.now(),
        'expires': datetime.datetime.now(),
        'files': [{
            'path': 'tests/fixtures/example.vcf',
            'archive': False,
            'tags': ['vcf', 'sample']
        }, {
            'path': 'tests/fixtures/family.vcf',
            'archive': True,
            'tags': ['vcf', 'family']
        }]
    }
    return data


@pytest.fixture(scope='function')
def bundle_data_old():
    data = {
        'name': 'angrybird',
        'created': dateutil.parser.parse('2018/01/01 00:00:00'),
        'expires': datetime.datetime.now(),
        'files': [{
            'path': 'tests/fixtures/example.2.vcf',
            'archive': False,
            'tags': ['vcf', 'sample']
        }, {
            'path': 'tests/fixtures/family.2.vcf',
            'archive': True,
            'tags': ['vcf', 'family']
        }]
    }
    return data


@pytest.yield_fixture(scope='function')
def store(tmpdir):
    _store = Store(uri='sqlite://', root=str(tmpdir))
    _store.create_all()
    yield _store
    _store.drop_all()



from functools import partial

from click.testing import CliRunner
import pytest

from housekeeper.cli import base


@pytest.fixture
def cli_runner():
    runner = CliRunner()
    return runner


@pytest.fixture
def invoke_cli(cli_runner):
    return partial(cli_runner.invoke, base)



# -*- coding: utf-8 -*-
from copy import deepcopy
import datetime


def test_build_tags(store):
    # GIVEN a store with some tags
    tag_names = ['vcf', 'family']
    new_tag = store.new_tag(tag_names[0])
    store.add_commit(new_tag)
    assert store.Tag.query.count() == 1
    assert store.Tag.query.first() == new_tag
    # WHEN adding tags...
    tag_map = store._build_tags(tag_names)
    # THEN it should return the existing tag
    assert len(tag_map) == len(tag_names)
    assert tag_map[tag_names[0]].name == tag_names[0]
    # ... and build the missing tag
    assert tag_map[tag_names[1]].name == tag_names[1]
    # which can be committed to the database
    assert store.Tag.query.count() == 1
    store.add_commit(list(tag_map.values()))
    assert store.Tag.query.count() == 2


def test_add_bundle(store, bundle_data):
    # GIVEN some input data for a new bundle
    assert store.Bundle.query.count() == 0
    # WHEN adding the new bundle
    bundle_obj, version_obj = store.add_bundle(bundle_data)
    # THEN it should look as expected
    assert bundle_obj.name == bundle_data['name']
    assert bundle_obj.created_at == bundle_data['created']
    assert bundle_obj.versions[0].created_at == bundle_data['created']
    assert bundle_obj.versions[0].expires_at == bundle_data['expires']
    assert len(bundle_obj.versions[0].files) == len(bundle_data['files'])
    # ... and you should be able to commit the record in one go!
    store.add_commit(bundle_obj)
    assert store.Bundle.query.count() == 1
    assert store.Bundle.query.first() == bundle_obj
    assert store.Tag.query.count() == 3
    assert store.File.query.count() == 2


def test_add_bundle_twice(store, bundle_data):
    # GIVEN a bundle in the store
    bundle_obj, version_obj = store.add_bundle(bundle_data)
    store.add_commit(bundle_obj)
    assert store.Bundle.query.first() == bundle_obj
    # WHEN adding the same bundle again
    # THEN it should return None
    new_bundle = store.add_bundle(bundle_data)
    assert new_bundle is None
    assert store.Bundle.query.count() == 1


def test_add_two_versions_of_bundle(store, bundle_data):
    # GIVEN two versions of the same bundle
    assert store.Bundle.query.count() == 0
    bundle_1 = bundle_data
    bundle_2 = deepcopy(bundle_data)
    bundle_2['created'] = datetime.datetime.now()
    for file_data in bundle_2['files']:
        file_data['path'] = file_data['path'].replace('.vcf', '.2.vcf')
    # WHEN adding them to the database
    with store.session.no_autoflush:
        for bundle in [bundle_1, bundle_2]:
            bundle_obj, version_obj = store.add_bundle(bundle)
            store.add_commit(bundle_obj)

    # THEN it should only create a single bundle
    assert store.Bundle.query.count() == 1
    # ... but two versions
    assert store.Version.query.count() == 2
    # ... and all four files
    assert store.File.query.count() == 4



# -*- coding: utf-8 -*-
import os

import coloredlogs
from flask import Flask
from flask_cors import CORS

import housekeeper
from housekeeper.server import api, ext

coloredlogs.install(level='INFO')
app = Flask(__name__)

SECRET_KEY = 'unsafe!!!'
TEMPLATES_AUTO_RELOAD = True
SQLALCHEMY_DATABASE_URI = os.environ['SQLALCHEMY_DATABASE_URI']
SQLALCHEMY_POOL_RECYCLE = 7200
SQLALCHEMY_TRACK_MODIFICATIONS = 'FLASK_DEBUG' in os.environ

app.config.from_object(__name__)

# register blueprints
app.register_blueprint(api.blueprint)

# configure extensions
ext.store.init_app(app)
cors = CORS(app, resources={r"/api/*": {'origins': '*'}})


@app.route('/')
def index():
    return f"Welcome to {housekeeper.__title__} REST API"



# -*- coding: utf-8 -*-
import datetime as dt
import logging
from typing import List
from pathlib import Path
from dateutil.parser import parse as parse_date

import alchy
from sqlalchemy import func

from housekeeper.add.core import AddHandler
from . import models

log = logging.getLogger(__name__)


class BaseHandler:

    Bundle = models.Bundle
    Version = models.Version
    File = models.File
    Tag = models.Tag

    def bundles(self):
        """Fetch bundles."""
        return self.Bundle.query

    def bundle(self, name: str) -> models.Bundle:
        """Fetch a bundle from the store."""
        return self.Bundle.filter_by(name=name).first()

    def version(self, bundle: str, date: dt.datetime) -> models.Version:
        """Fetch a version from the store."""
        return (self.Version.query
                            .join(models.Version.bundle)
                            .filter(models.Bundle.name == bundle,
                                    models.Version.created_at == date)
                            .first())

    def tag(self, name: str) -> models.Tag:
        """Fetch a tag from the database."""
        return self.Tag.filter_by(name=name).first()

    def new_bundle(self, name: str, created_at: dt.datetime=None) -> models.Bundle:
        """Create a new file bundle."""
        new_bundle = self.Bundle(name=name, created_at=created_at)
        return new_bundle

    def new_version(self, created_at: dt.datetime, expires_at: dt.datetime=None) -> models.Version:
        """Create a new bundle version."""
        new_version = self.Version(created_at=created_at, expires_at=expires_at)
        return new_version

    def new_file(self, path: str, checksum: str=None, to_archive: bool=False,
                 tags: List[models.Tag]=None) -> models.File:
        """Create a new file."""
        new_file = self.File(path=path, checksum=checksum, to_archive=to_archive, tags=tags)
        return new_file

    def new_tag(self, name: str, category: str=None) -> models.Tag:
        """Create a new tag."""
        new_tag = self.Tag(name=name, category=category)
        return new_tag

    def file_(self, file_id: int):
        """Get a file by record id."""
        return self.File.get(file_id)

    def files(self, *, bundle: str=None, tags: List[str]=None, version: int=None,
              path: str=None) -> models.File:
        """Fetch files from the store."""
        query = self.File.query
        if bundle:
            query = (query.join(self.File.version, self.Version.bundle)
                          .filter(self.Bundle.name == bundle))

        if tags:
            # require records to match ALL tags
            query = (
                query.join(self.File.tags)
                .filter(self.Tag.name.in_(tags))
                .group_by(models.File.id)
                .having(func.count(models.Tag.name) == len(tags))
            )

        if version:
            query = query.join(self.File.version).filter(self.Version.id == version)

        if path:
            query = query.filter_by(path=path)

        return query


    def files_before(self, *, bundle: str=None, tags: List[str]=None, before:
                     str=None) -> models.File:
        """Fetch files before date from store"""
        query = self.files(tags=tags, bundle=bundle)
        if before:
            before_dt = parse_date(before)
            query = query.join(models.Version).filter(models.Version.created_at < before_dt)

        return query


    def files_ondisk(self, file_objs: models.File) -> set:
        """Returns a list of files that are not on disk."""

        return set([ file_obj for file_obj in file_objs if Path(file_obj.full_path).is_file() ])


class Store(alchy.Manager, BaseHandler, AddHandler):

    """
    Handles interactions with the database in the context when a temporary
    database connection is needed, e.g. a command line interface.

    Args:
        uri (str): SQLAlchemy database connection str
    """

    def __init__(self, uri: str, root: str):
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=models.Model)
        self.File.app_root = Path(root)
        self.Version.app_root = Path(root)



# -*- coding: utf-8 -*-
from pathlib import Path
import datetime

from housekeeper.store import models


def test_Version():
    # GIVEN a bundle version
    bundle_obj = models.Bundle(name='handsomepig')
    now = datetime.datetime.now()
    version_obj = models.Version(created_at=now, bundle=bundle_obj)
    # WHEN accessing the relative root path
    root_dir = version_obj.relative_root_dir
    # THEN it should point to the correct folder
    now_str = str(now.date())
    assert root_dir == Path('handsomepig') / now_str

def test_Version(tmpdir, store):
    # GIVEN a bundle version
    bundle_obj = models.Bundle(name='handsomepig')
    now = datetime.datetime.now()
    version_obj = models.Version(created_at=now, bundle=bundle_obj)
    # WHEN accessing the full path
    root_dir = version_obj.full_path
    # THEN it should point to the correct folder
    now_str = str(now.date())
    assert root_dir == Path(tmpdir) / 'handsomepig' / now_str



# -*- coding: utf-8 -*-
from marshmallow import Schema, fields


class FileSchema(Schema):
    path = fields.Str(required=True, error_messages={'required': 'File path is required.'})
    archive = fields.Boolean(default=False)
    tags = fields.List(fields.Str(), required=True)


class BundleSchema(Schema):
    name = fields.Str(required=True, error_messages={'required': 'Bundle name is required.'})
    created = fields.DateTime(required=True,
                              error_messages={'required': 'Bundle date is required.'})
    expires = fields.DateTime()
    files = fields.List(fields.Nested(FileSchema), required=True)



# -*- coding: utf-8 -*-
import logging
from pathlib import Path
from typing import List

from housekeeper.store import models

LOG = logging.getLogger(__name__)


class AddHandler:

    def add_bundle(self, data: dict) -> models.Bundle:
        """Build a new bundle version of files.

        The format of the input dict is defined in the `schema` module.
        """
        bundle_obj = self.bundle(data['name'])
        if bundle_obj and self.version(bundle_obj.name, data['created']):
            LOG.debug('version of bundle already added')
            return None

        if bundle_obj is None:
            bundle_obj = self.new_bundle(name=data['name'], created_at=data['created'])

        version_obj = self.new_version(created_at=data['created'], expires_at=data.get('expires'))

        tag_names = set(tag_name for file_data in data['files'] for tag_name in file_data['tags'])
        tag_map = self._build_tags(tag_names)

        for file_data in data['files']:
            LOG.debug(f"adding file: {file_data['path']}")
            if not Path(file_data['path']).exists():
                raise FileNotFoundError(file_data['path'])
            tags = [tag_map[tag_name] for tag_name in file_data['tags']]
            new_file = self.new_file(file_data['path'], to_archive=file_data['archive'], tags=tags)
            version_obj.files.append(new_file)

        version_obj.bundle = bundle_obj
        return bundle_obj, version_obj

    def _build_tags(self, tag_names: List[str]) -> dict:
        """Build a list of tag objects."""
        tags = {}
        for tag_name in tag_names:
            tag_obj = self.tag(tag_name)
            if tag_obj is None:
                LOG.debug(f"create new tag: {tag_name}")
                tag_obj = self.new_tag(tag_name)
            tags[tag_name] = tag_obj
        return tags



# -*- coding: utf-8 -*-
from pathlib import Path
from typing import List

import click

from housekeeper.store import Store


@click.group()
@click.pass_context
def add(context):
    """Add things to the store."""
    context.obj['db'] = Store(context.obj['database'], context.obj['root'])


@add.command()
@click.argument('name')
@click.pass_context
def bundle(context, name):
    """Add a new bundle."""
    if context.obj['db'].bundle(name):
        click.echo(click.style('bundle name already exists', fg='yellow'))
        context.abort()
    new_bundle = context.obj['db'].new_bundle(name)
    context.obj['db'].add_commit(new_bundle)

    # add default version
    new_version = context.obj['db'].new_version(created_at=new_bundle.created_at)
    new_version.bundle = new_bundle
    context.obj['db'].add_commit(new_version)

    click.echo(click.style(f"new bundle added: {new_bundle.name} ({new_bundle.id})", fg='green'))


@add.command('file')
@click.option('-t', '--tag', 'tags', multiple=True, help='tag to associate the file by')
@click.option('-a', '--archive', is_flag=True, help='mark file to be archived')
@click.argument('bundle_name')
@click.argument('path')
@click.pass_context
def file_cmd(context, tags, archive, bundle_name, path):
    """Add a file to a bundle."""
    bundle_obj = context.obj['db'].bundle(bundle_name)
    if bundle_obj is None:
        click.echo(click.style(f"unknown bundle: {bundle_name}", fg='red'))
        context.abort()
    version_obj = bundle_obj.versions[0]
    new_file = context.obj['db'].new_file(
        path=str(Path(path).absolute()),
        to_archive=archive,
        tags=[context.obj['db'].tag(tag_name) if context.obj['db'].tag(tag_name) else
              context.obj['db'].new_tag(tag_name) for tag_name in tags]
    )
    new_file.version = version_obj
    context.obj['db'].add_commit(new_file)
    click.echo(click.style(f"new file added: {new_file.path} ({new_file.id})", fg='green'))


@add.command()
@click.argument('file_id', type=int)
@click.argument('tags', nargs=-1)
@click.pass_context
def tag(context: click.Context, file_id: int, tags: List[str]):
    """Add tags to an existing file."""
    file_obj = context.obj['db'].file_(file_id)
    if file_obj is None:
        print(click.style('unable to find file', fg='red'))
        context.abort()
    for tag_name in tags:
        tag_obj = context.obj['db'].tag(tag_name)
        if tag_obj is None:
            tag_obj = context.obj['db'].new_tag(tag_name)
        elif tag_obj in file_obj.tags:
            print(click.style(f"{tag_name}: tag already added", fg='yellow'))
            continue
        file_obj.tags.append(tag_obj)
    context.obj['db'].commit()
    all_tags = (tag.name for tag in file_obj.tags)
    print(click.style(f"file tags: {', '.join(all_tags)}", fg='blue'))



# -*- coding: utf-8 -*-
import click
import coloredlogs
import ruamel.yaml

import housekeeper
from . import add, include, init, get, delete


@click.group()
@click.option('-c', '--config', type=click.File())
@click.option('-d', '--database', help='path/URI of the SQL database')
@click.option('-r', '--root', type=click.Path(exists=True), help='Housekeeper root dir')
@click.option('-l', '--log-level', default='INFO')
@click.version_option(housekeeper.__version__, prog_name=housekeeper.__title__)
@click.pass_context
def base(context, config, database, root, log_level):
    """Housekeeper - Access your files!"""
    coloredlogs.install(level=log_level)
    context.obj = ruamel.yaml.safe_load(config) if config else {}
    context.obj['database'] = database if database else context.obj['database']
    context.obj['root'] = root if root else context.obj['root']


base.add_command(init.init)
base.add_command(add.add)
base.add_command(include.include)
base.add_command(get.get)
base.add_command(delete.delete)



# -*- coding: utf-8 -*-
from copy import deepcopy
import datetime
import pytest

import tempfile


@pytest.fixture(scope='function')
def bundle_data_json(bundle_data):
    json_data = deepcopy(bundle_data)
    json_data['created'] = json_data['created'].isoformat()
    json_data['expires'] = json_data['expires'].isoformat()
    return json_data



# -*- coding: utf-8 -*-
from flask import Blueprint, jsonify, request

from housekeeper.server.ext import store

blueprint = Blueprint('api', __name__, url_prefix='/api/v1')


@blueprint.route('/bundles')
def bundles():
    """Display bundles."""
    per_page = int(request.args.get('per_page', 30))
    page = int(request.args.get('page', 1))
    query = store.bundles()

    query_page = query.paginate(page, per_page=per_page)
    data = []
    for bundle_obj in query_page.items:
        bundle_data = bundle_obj.to_dict()
        bundle_data['versions'] = [version.to_dict() for version in bundle_obj.versions]
        data.append(bundle_data)

    return jsonify(bundles=data)



# -*- coding: utf-8 -*-
import datetime
from pathlib import Path

import pytest

from housekeeper import include
from housekeeper.exc import VersionIncludedError


def test_checksum():
    # GIVEN a file with a specific checksum
    file_path = Path('tests/fixtures/26a90105b99c05381328317f913e9509e373b64f.txt')
    expected_checksum = file_path.name.rstrip('.txt')
    # WHEN calculating the checksum
    checksum = include.checksum(file_path)
    # THEN it should match
    assert checksum == expected_checksum


def test_include_version(tmpdir, version):
    # GIVEN a new directory as the global root
    global_root = tmpdir.mkdir('destination')
    global_root_path = Path(global_root)
    file_names = [Path(file_obj.path).name for file_obj in version.files]
    date_str = str(version.created_at.date())
    assert len(list(global_root_path.iterdir())) == 0
    # WHEN including a bundle version
    include.include_version(global_root, version)
    # THEN it should link the files to the new destination
    version_root_path = global_root_path / version.bundle.name / date_str
    assert len(list(version_root_path.iterdir())) == len(version.files)
    for file_obj, file_name in zip(version.files, file_names):
        # ... and the file path should be replace with relative "internal" path
        assert (global_root_path / file_obj.path).exists()
        assert (global_root_path / file_obj.path).is_file()
        assert Path(file_obj.path) == Path(version.bundle.name) / date_str / file_name
        # ... and it should fill-in the checksum
        if file_obj.to_archive:
            assert isinstance(file_obj.checksum, str)
        else:
            assert file_obj.checksum is None


def test_include_version_already_included(tmpdir, version):
    # GIVEN a version which is already marked as included
    version.included_at = datetime.datetime.now()
    # WHEN trying to "include it"
    # THEN it should raise an exception
    with pytest.raises(VersionIncludedError):
        include.include_version(tmpdir, version)



from typing import List

import click

from housekeeper.store import Store


@click.command()
@click.option('-t', '--tag', 'tags', multiple=True, help='filter by file tag')
@click.option('-v', '--version', type=int, help='filter by version of the bundle')
@click.option('-V', '--verbose', is_flag=True, help='print additional information')
@click.argument('bundle', required=False)
@click.pass_context
def get(context, tags: List[str], version: int, verbose: bool, bundle: str):
    """Get files."""
    store = Store(context.obj['database'], context.obj['root'])
    files = store.files(bundle=bundle, tags=tags, version=version)
    for file_obj in files:
        if verbose:
            tags = ', '.join(tag.name for tag in file_obj.tags)
            click.echo(f"{click.style(str(file_obj.id), fg='blue')} | {file_obj.full_path} | "
                       f"{click.style(tags, fg='yellow')}")
        else:
            click.echo(file_obj.full_path)



# -*- coding: utf-8 -*-
from housekeeper.add import schema


def test_schema_with_valid_input(bundle_data_json):
    # GIVEN valid input for a new bundle
    # WHEN validating it with the schema specs
    errors = schema.BundleSchema().validate(bundle_data_json)
    # THEN it should not report any errors
    assert errors == {}


def test_schema_with_invalid_input(bundle_data_json):
    # GIVEN input data with missing name of the bundle
    del bundle_data_json['name']
    # WHEN validating it against the schema
    errors = schema.BundleSchema().validate(bundle_data_json)
    # THEN it should report errors for the field
    assert len(errors) > 0
    assert 'name' in errors



# -*- coding: utf-8 -*-
import hashlib
import logging
import os
from pathlib import Path

from housekeeper.exc import VersionIncludedError
from housekeeper.store import models

BLOCKSIZE = 65536
log = logging.getLogger(__name__)


def include_version(global_root: str, version_obj: models.Version, hardlink:bool=True):
    """Include files in existing bundle version."""
    global_root_dir = Path(global_root)
    if version_obj.included_at:
        raise VersionIncludedError(f"version included on {version_obj.included_at}")

    # generate root directory
    version_root_dir = global_root_dir / version_obj.relative_root_dir
    version_root_dir.mkdir(parents=True, exist_ok=True)
    log.info(f"created new bundle version dir: {version_root_dir}")

    for file_obj in version_obj.files:
        # hardlink file to the internal structure
        file_obj_path = Path(file_obj.path)
        new_path = version_root_dir / file_obj_path.name
        if hardlink:
            os.link(file_obj_path.resolve(), new_path)
        else:
            os.symlink(file_obj_path.resolve(), new_path)
        log.info(f"linked file: {file_obj.path} -> {new_path}")
        file_obj.path = str(new_path).replace(f"{global_root_dir}/", '', 1)


def checksum(path):
    """Calculcate checksum for a file."""
    hasher = hashlib.sha1()
    with open(path, 'rb') as stream:
        buf = stream.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = stream.read(BLOCKSIZE)
    return hasher.hexdigest()



# -*- coding: utf-8 -*-
from pathlib import Path

import ruamel.yaml

import housekeeper
from housekeeper.store import Store


def test_cli_version(invoke_cli):
    # GIVEN I want to see the version of the program
    # WHEN asking to see the version
    result = invoke_cli(['--version'])
    # THEN it should display the version of the program
    # THEN it should print the name and version of the tool only
    assert housekeeper.__title__ in result.output
    assert housekeeper.__version__ in result.output


def test_cli_config(tmpdir, invoke_cli):
    # GIVEN a simple, valid cli
    config_file = tmpdir.join('config.yaml')
    config_file.write(ruamel.yaml.safe_dump(dict(root=str(tmpdir), database='sqlite://')))
    # WHEN calling the CLI
    result = invoke_cli(['--config', str(config_file), 'init', '--help'])
    # THEN it should read the config and import it to the context
    assert result.exit_code == 0
    # ... not sure how to test :-( ...


def test_cli_init(cli_runner, invoke_cli):
    # GIVEN you want to setup a new database using the CLI
    database = './test_db.sqlite3'
    database_path = Path(database)
    database_uri = f"sqlite:///{database}"
    with cli_runner.isolated_filesystem():
        assert database_path.exists() is False

        # WHEN calling "init"
        result = invoke_cli(['--database', database_uri, '--root', '.', 'init'])

        # THEN it should setup the database with some tables
        assert result.exit_code == 0
        assert database_path.exists()
        assert len(Store(database_uri, root='/tmp').engine.table_names()) > 0

        # GIVEN the database already exists
        # WHEN calling the init function
        result = invoke_cli(['--database', database_uri, '--root', '.', 'init'])
        # THEN it should print an error and give error exit code
        assert result.exit_code != 0
        assert 'Database already exists' in result.output

        # GIVEN the database already exists
        # WHEN calling "init" with "--reset"
        result = invoke_cli(['--database', database_uri, '--root', '.', 'init', '--reset'],
                            input='Yes')
        # THEN it should re-setup the tables and print new tables
        assert result.exit_code == 0
        assert 'Success!' in result.output


def test_cli_delete_files(tmpdir, invoke_cli, bundle_data):
    config_file = tmpdir.join('config.yaml')
    database_path = Path(tmpdir).joinpath('database.sqlite')
    config_file.write(ruamel.yaml.safe_dump(dict(root=str(tmpdir), database='sqlite:///' + str(database_path))))
    store = Store(uri='sqlite:///' + str(database_path), root=str(tmpdir))
    store.create_all()

    # GIVEN you want to remove everything
    result = invoke_cli(['--config', str(config_file), 'delete', 'files'])
    # THEN HAL9000 should interfere
    assert result.exit_code == 1
    assert result.output == "I'm afraid I can't let you do that.\nAborted!\n"

    # GIVEN you want to delete a not existing bundle
    result = invoke_cli(['--config', str(config_file),'delete', 'files', '--bundle', 'sillyfish'])
    # THEN it should not be found
    assert result.exit_code == 1
    assert result.output == "bundle not found\nAborted!\n"

    # GIVEN you want to delete a bundle
    bundle_obj, version_obj = store.add_bundle(data=bundle_data)
    store.add_commit(bundle_obj)
    result = invoke_cli(['--config', str(config_file),'delete', 'files', '--bundle', 'sillyfish'])
    # THEN it should ask you if you are sure
    assert result.exit_code == 1
    assert result.output == "Are you sure you want to delete 2 files? [y/N]: \nAborted!\n"

    store.drop_all()



# -*- coding: utf-8 -*-
import click

from housekeeper.store import Store


@click.command()
@click.option('--reset', is_flag=True, help='reset database before setting up tables')
@click.option('--force', is_flag=True, help='bypass manual confirmations')
@click.pass_context
def init(context, reset, force):
    """Setup the database."""
    store = Store(context.obj['database'], context.obj['root'])
    existing_tables = store.engine.table_names()
    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg='yellow'), abort=True)
        store.drop_all()
    elif existing_tables:
        click.echo(click.style("Database already exists, use '--reset'", fg='red'))
        context.abort()

    store.create_all()
    message = f"Success! New tables: {', '.join(store.engine.table_names())}"
    click.echo(click.style(message, fg='green'))



# -*- coding: utf-8 -*-
import datetime as dt
from pathlib import Path

import alchy
from sqlalchemy import Column, ForeignKey, orm, types, UniqueConstraint, Table

Model = alchy.make_declarative_base(Base=alchy.ModelBase)


file_tag_link = Table(
    'file_tag_link',
    Model.metadata,
    Column('file_id', types.Integer, ForeignKey('file.id'), nullable=False),
    Column('tag_id', types.Integer, ForeignKey('tag.id'), nullable=False),
    UniqueConstraint('file_id', 'tag_id', name='_file_tag_uc'),
)


class Bundle(Model):

    """A general group of files."""

    __tablename__ = 'bundle'

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(64), unique=True, nullable=False)
    created_at = Column(types.DateTime, default=dt.datetime.now)
    versions = orm.relationship('Version', backref='bundle', order_by='-Version.created_at', cascade='delete, save-update')


class Version(Model):

    __tablename__ = 'version'

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, nullable=False)
    expires_at = Column(types.DateTime)
    included_at = Column(types.DateTime)
    removed_at = Column(types.DateTime)

    archived_at = Column(types.DateTime)
    archive_path = Column(types.String(256), unique=True)
    archive_checksum = Column(types.String(256), unique=True)

    bundle_id = Column(ForeignKey(Bundle.id, ondelete='CASCADE'), nullable=False)
    files = orm.relationship('File', backref='version', cascade='delete, save-update')

    app_root = None

    @property
    def relative_root_dir(self):
        """Build the relative root dir path for the bundle version."""
        return Path(self.bundle.name) / str(self.created_at.date())

    @property
    def full_path(self):
        return Path(self.app_root) / self.bundle.name / str(self.created_at.date())


class File(Model):

    """Represent a file."""

    __tablename__ = 'file'

    id = Column(types.Integer, primary_key=True)
    path = Column(types.String(256), unique=True, nullable=False)
    checksum = Column(types.String(256))
    to_archive = Column(types.Boolean, nullable=False, default=False)

    version_id = Column(ForeignKey(Version.id, ondelete='CASCADE'), nullable=False)
    tags = orm.relationship('Tag', secondary=file_tag_link, backref='files')

    app_root = None

    @property
    def full_path(self):
        """Return the full path to the file."""
        if Path(self.path).is_absolute():
            return self.path
        else:
            return str(self.app_root / self.path)

    @property
    def is_included(self):
        """Check if the file is included in Housekeeper."""
        return str(self.app_root) in self.full_path


class Tag(Model):

    __tablename__ = 'tag'

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(64), unique=True, nullable=False)
    category = Column(types.String(64))
    created_at = Column(types.DateTime, default=dt.datetime.now)



from pathlib import Path
import shutil

import click

from housekeeper.store import Store, models


@click.group()
@click.pass_context
def delete(context):
    """Delete things in the database."""
    context.obj['store'] = Store(context.obj['database'], context.obj['root'])


@delete.command()
@click.option('-y', '--yes', is_flag=True, help='skip checks')
@click.argument('bundle_name')
@click.pass_context
def bundle(context, yes, bundle_name):
    """Delete the latest bundle version."""
    bundle_obj = context.obj['store'].bundle(bundle_name)
    if bundle_obj is None:
        click.echo(click.style('bundle not found', fg='red'))
        context.abort()
    version_obj = bundle_obj.versions[0]
    if version_obj.included_at:
        question = f"remove bundle version from file system and database: {version_obj.full_path}"
    else:
        question = f"remove bundle version from database: {version_obj.created_at.date()}"
    if yes or click.confirm(question):
        if version_obj.included_at:
            shutil.rmtree(version_obj.full_path, ignore_errors=True)
        version_obj.delete()
        context.obj['store'].commit()
        click.echo(f"version deleted: {version_obj.full_path}")


@delete.command()
@click.option('--yes', multiple=True, is_flag=True, help='skip checks')
@click.option('-t', '--tag', multiple=True, help='file tag')
@click.option('-b', '--bundle', help='bundle name')
@click.option('-a', '--before', help='version created before...')
@click.option('-n', '--notondisk', is_flag=True, help='rm db entry from files not on disk')
@click.pass_context
def files(context, yes, tag, bundle, before, notondisk):
    """Delete files based on tags."""
    file_objs = []

    if not tag and not bundle:
        click.echo("I'm afraid I can't let you do that.")
        context.abort()

    if bundle:
        bundle_obj = context.obj['store'].bundle(bundle)
        if bundle_obj is None:
            click.echo(click.style('bundle not found', fg='red'))
            context.abort()

    query = context.obj['store'].files_before(bundle = bundle, tags = tag, before = before)

    if notondisk:
        file_objs = set(query) - context.obj['store'].files_ondisk(query)
    else:
        file_objs = query.all()

    if len(file_objs) > 0 and len(yes) < 2:
        if not click.confirm(f"Are you sure you want to delete {len(file_objs)} files?"):
            context.abort()

    for file_obj in file_objs:
        if yes or click.confirm(f"remove file from disk and database: {file_obj.full_path}"):
            file_obj_path = Path(file_obj.full_path)
            if file_obj.is_included and (file_obj_path.exists() or file_obj_path.is_symlink()):
                file_obj_path.unlink()
            file_obj.delete()
            context.obj['store'].commit()
            click.echo(f'{file_obj.full_path} deleted')


@delete.command('file')
@click.option('-y', '--yes', is_flag=True, help='skip checks')
@click.argument('file_id', type=int)
@click.pass_context
def file_cmd(context, yes, file_id):
    """Delete a file."""
    file_obj = context.obj['store'].File.get(file_id)
    if file_obj.is_included:
        question = f"remove file from file system and database: {file_obj.full_path}"
    else:
        question = f"remove file from database: {file_obj.full_path}"
    if yes or click.confirm(question):
        if file_obj.is_included and Path(file_obj.full_path).exists():
            Path(file_obj.full_path).unlink()

        file_obj.delete()
        context.obj['store'].commit()
        click.echo('file deleted')



# -*- coding: utf-8 -*-
import datetime as dt

import click

from housekeeper.exc import VersionIncludedError
from housekeeper.include import include_version
from housekeeper.store import Store


@click.command()
@click.option('-v', '--version', type=int, help='version id of the bundle version')
@click.argument('bundle_name', required=False)
@click.pass_context
def include(context, bundle_name, version):
    """Include a bundle of files into the internal space.

    Use bundle name if you simply want to inlcude the latest version.
    """
    store = Store(context.obj['database'], context.obj['root'])
    if version:
        version_obj = store.Version.get(version)
        if version_obj is None:
            click.echo(click.style('version not found', fg='red'))
    else:
        bundle_obj = store.bundle(bundle_name)
        if bundle_obj is None:
            click.echo(click.style('bundle not found', fg='red'))
        version_obj = bundle_obj.versions[0]

    try:
        include_version(context.obj['root'], version_obj)
    except VersionIncludedError as error:
        click.echo(click.style(error.message, fg='red'))
        context.abort()

    version_obj.included_at = dt.datetime.now()
    store.commit()
    click.echo(click.style('included all files!', fg='green'))
